sound_type snd_make_sax(double freq, sound_type breath_env, rate_type sr);
sound_type snd_sax(double freq, sound_type breath_env, rate_type sr);
    /* LISP: (snd-sax ANYNUM SOUND ANYNUM) */
#define SAX_CONTROL_CHANGE_CONST 128
