sound_type snd_make_clarinet(double freq, sound_type breath_env, rate_type sr);
sound_type snd_clarinet(double freq, sound_type breath_env, rate_type sr);
    /* LISP: (snd-clarinet ANYNUM SOUND ANYNUM) */
#define CLAR_CONTROL_CHANGE_CONST 128
