sound_type snd_make_pluck(rate_type sr, double hz, time_type t0, time_type d, double final_amp);
sound_type snd_pluck(rate_type sr, double hz, time_type t0, time_type d, double final_amp);
    /* LISP: (snd-pluck ANYNUM ANYNUM ANYNUM ANYNUM ANYNUM) */
