sound_type snd_make_resoncv(sound_type s1, double hz, sound_type bw, int normalization);
sound_type snd_resoncv(sound_type s1, double hz, sound_type bw, int normalization);
    /* LISP: (snd-resoncv SOUND ANYNUM SOUND LONG) */
