sound_type snd_make_tapv(sound_type s1, double offset, sound_type vardelay, double maxdelay);
sound_type snd_tapv(sound_type s1, double offset, sound_type vardelay, double maxdelay);
    /* LISP: (snd-tapv SOUND ANYNUM SOUND ANYNUM) */
