// localdefs.h -- additional declarations needed for XLisp extensions
//                this file is for Nyquist

/* include actual local file headers: */
#include "sndfnintdefs.h"
#include "seqfnintdefs.h"

/* from sndsliders.c */

LVAL xslider_read(void);


