sound_type snd_make_osc(sound_type input, double step, rate_type sr, double hz, time_type t0, time_type d, double phase);
sound_type snd_osc(sound_type input, double step, rate_type sr, double hz, time_type t0, time_type d, double phase);
    /* LISP: (snd-osc SOUND ANYNUM ANYNUM ANYNUM ANYNUM ANYNUM ANYNUM) */
