sound_type snd_make_lpreson(sound_type x_snd, LVAL src, time_type frame_time);
sound_type snd_lpreson(sound_type x_snd, LVAL src, time_type frame_time);
    /* LISP: (snd-lpreson SOUND ANY ANYNUM) */
