extern LVAL xlc_snd_set_max_audio_mem(void);
extern LVAL xlc_snd_set_latency(void);
extern LVAL xlc_soundp(void);
extern LVAL xlc_hz_to_step(void);
extern LVAL xlc_snd_set_logical_stop(void);
extern LVAL xlc_log(void);
extern LVAL xlc_snd_sref(void);
extern LVAL xlc_sref_inverse(void);
extern LVAL xlc_snd_stop_time(void);
extern LVAL xlc_snd_time(void);
extern LVAL xlc_snd_srate(void);
extern LVAL xlc_snd_t0(void);
extern LVAL xlc_snd_xform(void);
extern LVAL xlc_block_watch(void);
extern LVAL xlc_sound_nth_block(void);
extern LVAL xlc_snd_copy(void);
extern LVAL xlc_snd_print(void);
extern LVAL xlc_snd_play(void);
extern LVAL xlc_stats(void);
extern LVAL xlc_snd_print_tree(void);
extern LVAL xlc_snd_scale(void);
extern LVAL xlc_snd_zero(void);
extern LVAL xlc_step_to_hz(void);
extern LVAL xlc_snd_add(void);
extern LVAL xlc_snd_avg(void);
extern LVAL xlc_snd_compose(void);
extern LVAL xlc_snd_convolve(void);
extern LVAL xlc_snd_down(void);
extern LVAL xlc_snd_fft(void);
extern LVAL xlc_snd_inverse(void);
extern LVAL xlc_snd_lpanal(void);
extern LVAL xlc_snd_multiseq(void);
extern LVAL xlc_snd_phasevocoder(void);
extern LVAL xlc_snd_resample(void);
extern LVAL xlc_snd_resamplev(void);
extern LVAL xlc_snd_from_array(void);
extern LVAL xlc_snd_samples(void);
extern LVAL xlc_snd_length(void);
extern LVAL xlc_snd_maxsamp(void);
extern LVAL xlc_snd_fetch(void);
extern LVAL xlc_snd_fetch_array(void);
extern LVAL xlc_snd_max(void);
extern LVAL xlc_snd_read(void);
extern LVAL xlc_snd_seq(void);
extern LVAL xlc_snd_slider(void);
extern LVAL xlc_snd_stoponzero(void);
extern LVAL xlc_snd_trigger(void);
extern LVAL xlc_snd_yin(void);
extern LVAL xlc_snd_abs(void);
extern LVAL xlc_snd_allpoles(void);
extern LVAL xlc_snd_alpass(void);
extern LVAL xlc_snd_alpasscv(void);
extern LVAL xlc_snd_alpassvc(void);
extern LVAL xlc_snd_alpassvv(void);
extern LVAL xlc_snd_amosc(void);
extern LVAL xlc_snd_areson(void);
extern LVAL xlc_snd_aresoncv(void);
extern LVAL xlc_snd_aresonvc(void);
extern LVAL xlc_snd_aresonvv(void);
extern LVAL xlc_snd_atone(void);
extern LVAL xlc_snd_atonev(void);
extern LVAL xlc_snd_biquad(void);
extern LVAL xlc_snd_buzz(void);
extern LVAL xlc_snd_chase(void);
extern LVAL xlc_snd_clip(void);
extern LVAL xlc_snd_congen(void);
extern LVAL xlc_snd_const(void);
extern LVAL xlc_snd_coterm(void);
extern LVAL xlc_snd_delay(void);
extern LVAL xlc_snd_delaycv(void);
extern LVAL xlc_snd_eqbandvvv(void);
extern LVAL xlc_snd_exp(void);
extern LVAL xlc_snd_fmfb(void);
extern LVAL xlc_snd_fmfbv(void);
extern LVAL xlc_snd_fmosc(void);
extern LVAL xlc_snd_follow(void);
extern LVAL xlc_snd_fromarraystream(void);
extern LVAL xlc_snd_fromobject(void);
extern LVAL xlc_snd_gate(void);
extern LVAL xlc_snd_ifft(void);
extern LVAL xlc_snd_bandedwg(void);
extern LVAL xlc_snd_bowed(void);
extern LVAL xlc_snd_bowed_freq(void);
extern LVAL xlc_snd_clarinet(void);
extern LVAL xlc_snd_clarinet_all(void);
extern LVAL xlc_snd_clarinet_freq(void);
extern LVAL xlc_snd_flute(void);
extern LVAL xlc_snd_flute_all(void);
extern LVAL xlc_snd_flute_freq(void);
extern LVAL xlc_snd_mandolin(void);
extern LVAL xlc_snd_modalbar(void);
extern LVAL xlc_snd_sax(void);
extern LVAL xlc_snd_sax_all(void);
extern LVAL xlc_snd_sax_freq(void);
extern LVAL xlc_snd_sitar(void);
extern LVAL xlc_snd_integrate(void);
extern LVAL xlc_snd_log(void);
extern LVAL xlc_snd_lpreson(void);
extern LVAL xlc_snd_maxv(void);
extern LVAL xlc_snd_offset(void);
extern LVAL xlc_snd_oneshot(void);
extern LVAL xlc_snd_osc(void);
extern LVAL xlc_snd_partial(void);
extern LVAL xlc_snd_pluck(void);
extern LVAL xlc_snd_prod(void);
extern LVAL xlc_snd_pwl(void);
extern LVAL xlc_snd_quantize(void);
extern LVAL xlc_snd_recip(void);
extern LVAL xlc_snd_reson(void);
extern LVAL xlc_snd_resoncv(void);
extern LVAL xlc_snd_resonvc(void);
extern LVAL xlc_snd_resonvv(void);
extern LVAL xlc_snd_sampler(void);
extern LVAL xlc_snd_normalize(void);
extern LVAL xlc_snd_shape(void);
extern LVAL xlc_snd_sine(void);
extern LVAL xlc_snd_siosc(void);
extern LVAL xlc_snd_slope(void);
extern LVAL xlc_snd_sqrt(void);
extern LVAL xlc_snd_tapf(void);
extern LVAL xlc_snd_tapv(void);
extern LVAL xlc_snd_tone(void);
extern LVAL xlc_snd_tonev(void);
extern LVAL xlc_snd_up(void);
extern LVAL xlc_snd_white(void);
extern LVAL xlc_snd_stkpitshift(void);
extern LVAL xlc_snd_stkrev(void);
extern LVAL xlc_snd_stkchorus(void);
extern LVAL xlc_snd_save(void);
extern LVAL xlc_snd_overwrite(void);
