sound_type snd_make_modalbar(time_type t0, double freq, int preset, time_type dur, rate_type sr);
sound_type snd_modalbar(time_type t0, double freq, int preset, time_type dur, rate_type sr);
    /* LISP: (snd-modalbar ANYNUM ANYNUM LONG ANYNUM ANYNUM) */
