sound_type snd_make_reson(sound_type s, double hz, double bw, int normalization);
sound_type snd_reson(sound_type s, double hz, double bw, int normalization);
    /* LISP: (snd-reson SOUND ANYNUM ANYNUM LONG) */
