sound_type snd_make_sitar(time_type t0, double freq, time_type dur, rate_type sr);
sound_type snd_sitar(time_type t0, double freq, time_type dur, rate_type sr);
    /* LISP: (snd-sitar ANYNUM ANYNUM ANYNUM ANYNUM) */
